@@ -8,13 +8,12 @@ from __future__ import absolute_import
 
 import json
 import logging
-import threading
 from datetime import datetime, timedelta, tzinfo
 
 import boto
 from boto.dynamodb.item import Item
 from boto.exception import DynamoDBResponseError
-
+from boto.dynamodb.exceptions import DynamoDBConditionalCheckFailedError
 
 log = logging.getLogger(__name__)
 dblog = logging.getLogger(__name__+".database-access")
@@ -33,13 +32,6 @@ class SchemaError(Exception):
     is used with a ``range_key``.
     """
 
-#FIXME: should be raised when Amazons fails because of this instead of a Boto stuff
-class ThroughputError(Exception):
-    """Raised when requested throughput can not be allocated by
-    :py:meth:`~.ConnectionBorg.create_table`. It probably means that either read
-    or write is below Amazon's minimum of 5.
-    """
-
 
 class MaxRetriesExceededError(Exception):
     """Raised when a failed operation couldn't be completed after retrying
@@ -54,11 +46,17 @@ class OverwriteError(Exception):
     """
 
 
-class ExpectedValueError(Exception):
-    """Conditional write failure.
-    Raised when the expected_values argument to :meth:`DynamoDBModel.save`
-    doesn't match what is stored in the database (i.e. when somebody changed
-    the DB's version of your object behind your back).
+class ConflictError(Exception):
+    """Atomic edition failure.
+    Raised when an Item has been changed between the read and the write operation
+    and this has been forbid by the ``raise_on_conflict`` argument of
+    :meth:`DynamoDBModel.save` (i.e. when somebody changed the DB's version of
+    your object behind your back).
+    """
+
+
+class InvalidRegionError(Exception):
+    """Raised when ``set_region()`` is called with an invalid region name.
     """
 
 
@@ -126,7 +124,7 @@ def _get_proto_value(schema_type):
     return schema_type()
 
 
-def _get_default_value(schema_type, default=None):
+def _get_default_value(schema_type, default=None, name=''):
     """Return a default value matching schema_type or default if provided:
 
       - For datetime.datetime, it's NOW.
@@ -139,18 +137,19 @@ def _get_default_value(schema_type, default=None):
        - default was callable and required arguments
        - default or its return value is not an instance of schema_type
 
-    :param schema_type class object to instanciate
-    :param default default value. May be a value or a callable (functions, class, ...) It must *NOT* require an any argument and it's type must match schema_type
+    :param schema_type: class object to instanciate
+    :param default: default value. May be a value or a callable (functions, class, ...) It must *NOT* require an any argument and it's type must match schema_type
+    :param name: field name to use in exception string if applicable
 
     """
     if default is not None:
         # If default is callable(function, constructor, ...), try to dereference it
-        if hasattr(default, '__call__'):
+        if callable(default):
             # Might raise a "TypeError" if arguments were needed
             default = default()
         # Check default value consitency
         if not isinstance(default, schema_type):
-            raise TypeError("Expected default value of type {}, got: {}".format(schema_type, type(default)))
+            raise TypeError("Expected default value of type {}, got value: '{}' of type: {} for field: '{}'".format(schema_type, str(default), type(default), name))
         else:
             return default
 
@@ -238,7 +237,7 @@ def _dynamodb_to_python(schema_type, value):
         # the strings in the database to be UTC (+00:00) for now.
         # TODO Handle arbitrary timezones (with manual parsing).
         if value.endswith('Z'):
-            value = value[:-2] + '+00:00'
+            value = value[:-1] + '+00:00'
         return datetime.strptime(
             value, "%Y-%m-%dT%H:%M:%S.%f+00:00").replace(tzinfo=utc_tz)
 
@@ -261,28 +260,23 @@ class ConnectionBorg(object):
     _shared_state = {
         "_aws_access_key_id": None,
         "_aws_secret_access_key": None,
-        # {thread_id: connection} mapping
-        "_connections": {},
+        "_region": None,
+        "_connection": None,
     }
 
     def __init__(self):
         self.__dict__ = self._shared_state
 
     def _get_connection(self):
-        """Return the DynamoDB connection for the current thread, establishing
-        it if required.
+        """Return the DynamoDB connection for the mapper
         """
-        current_thread = threading.current_thread()
-        thread_id = current_thread.ident
-        try:
-            return self._connections[thread_id]
-        except KeyError:
-            log.debug("Creating DynamoDB connection for thread %s.", current_thread)
-            self._connections[thread_id] = boto.connect_dynamodb(
+        if self._connection is None:
+            self._connection = boto.connect_dynamodb(
                 aws_access_key_id=self._aws_access_key_id,
-                aws_secret_access_key=self._aws_secret_access_key
+                aws_secret_access_key=self._aws_secret_access_key,
+                region=self._region,
             )
-            return self._connections[thread_id]
+        return self._connection
 
     def _create_autoincrement_magic_item(self, table):
         item = table.new_item(hash_key=MAGIC_KEY, attrs={
@@ -305,6 +299,18 @@ class ConnectionBorg(object):
         self._aws_access_key_id = aws_access_key_id
         self._aws_secret_access_key = aws_secret_access_key
 
+    def set_region(self, region_name):
+        """Set the DynamoDB region. If this is not set AWS defaults to 'us-east-1'.
+
+        :param region_name: The name of the region to use
+        """
+        for region in boto.dynamodb.regions():
+            if region.name == region_name:
+                self._region = region
+                return
+
+        raise InvalidRegionError("Region name %s is invalid" % region_name)
+
     def create_table(self, cls, read_units, write_units, wait_for_active=False):
         """Create a table that'll be used to store instances of cls.
 
@@ -338,9 +344,6 @@ class ConnectionBorg(object):
         if not cls.__schema__:
             raise SchemaError("Class does not define __schema__", cls)
 
-        if read_units < 5 or write_units < 5:
-            raise ThroughputError("Either read or write throughput is below 5")
-
         hash_key_type = cls.__schema__[hash_key_name]
 
         if hash_key_type is autoincrement_int:
@@ -386,10 +389,6 @@ class ConnectionBorg(object):
         """Return the table with the requested name."""
         return self._get_connection().get_table(name)
 
-    def new_batch_list(self):
-        """Create a new batch list."""
-        return self._get_connection().new_batch_list()
-
 
 class DynamoDBModel(object):
     """Abstract base class for all models that use DynamoDB as their storage
@@ -412,10 +411,11 @@ class DynamoDBModel(object):
           corresponding to the type will be used.
           "defaulter" may either be a scalar value or a callable with no
           arguments.
+      - ``__migrator__``: :py:class:`~.Migration` handler attached to this model
 
     To redefine serialization/deserialization semantics (e.g. to have more
     complex schemas, like auto-serialized JSON data structures), override the
-    from_dict (deserialization) and to_db_dict (serialization) methods.
+    _from_dict (deserialization) and _to_db_dict (serialization) methods.
 
     *Important implementation note regarding sets:* DynamoDB can't store empty
     sets/strings. Therefore, since we have schema information available to us,
@@ -467,6 +467,7 @@ class DynamoDBModel(object):
     __hash_key__ = None
     __range_key__ = None
     __schema__ = None
+    __migrator__ = None
     __defaults__ = {}
 
     def __init__(self, **kwargs):
@@ -479,34 +480,60 @@ class DynamoDBModel(object):
 
         We're supplying this method to avoid the need for extra checks in save and
         ease object initial creation.
+
+        Objects created and initialized with this method are considered as not
+        coming from the DB.
         """
-        defaults = type(self).__defaults__
-        schema = type(self).__schema__
+        cls = type(self)
+        defaults = cls.__defaults__
+        schema = cls.__schema__
+
+        self._raw_data = {}
 
         for (name, type_) in schema.iteritems():
             if name in kwargs:
                 default = kwargs.get(name)
             else:
                 default = defaults[name] if name in defaults else None
-            setattr(self, name, _get_default_value(type_, default))
+            setattr(self, name, _get_default_value(type_, default, name=name))
+
+        # instanciate the migrator only once per model *after* initialization
+        if isinstance(cls.__migrator__, type):
+            cls.__migrator__ = cls.__migrator__(cls)
 
     @classmethod
-    def from_dict(cls, d):
-        """Build an instance from a dict-like mapping,
-        according to the class's schema.
+    def _from_db_dict(cls, raw_data):
+        """Build an instance from a dict-like mapping, according to the class's
+        schema. Objects created with this method are considered as comming from
+        the DB. The initial state is persisted in ``self._raw_data``.
+        If a ``__migrator__`` has been declared, migration is triggered on a copy
+        of the raw data.
 
         Default values are used for anything that's missing from the dict
         (see DynamoDBModel class docstring).
+
+        Direct use of this method should be avoided as much as possible but still
+        may be usefull for "deep copy".
+
+        Overload this method if you need a special (de-)serialization semantic
+
+        :param raw_data: Raw db dict
         """
-        #FIXME: type check. moving to __init__ syntax may break some impementations
-        #FIXME: rename to from_db_dict when breaking compat or disambiguing
+        #FIXME: type check. moving to __init__ syntax may break some implementations
         instance = cls()
+        instance._raw_data = raw_data
+
+        #If a migrator is registered, trigger it
+        if cls.__migrator__ is not None:
+           raw_data = cls.__migrator__(raw_data)
+
+        # de-serialize data
         for (name, type_) in cls.__schema__.iteritems():
             # Set the value if we got one from DynamoDB. Otherwise, stick with the default
-            value = _dynamodb_to_python(type_, d.get(name))
-            if value is None:
-                value = _get_default_value(type_)
+            value = _dynamodb_to_python(type_, raw_data.get(name)) # de-serialize
+            value = _get_default_value(type_, value, name=name)    # match value
             setattr(instance, name, value)
+
         return instance
 
     @classmethod
@@ -518,6 +545,9 @@ class DynamoDBModel(object):
         using a composite primary key, you need to specify both the hash and
         range key values.
 
+        Objects loaded by this method are marked as coming from the DB. Hence
+        their initial state is saved in ``self._raw_data``.
+
         :param hash_key_value: The value of the requested item's hash_key.
 
         :param range_key_value: The value of the requested item's range_key,
@@ -541,7 +571,7 @@ class DynamoDBModel(object):
 
         dblog.debug("Got item (%s, %s) from table %s", h_value, r_value, cls.__table__)
 
-        return cls.from_dict(item)
+        return cls._from_db_dict(item)
 
     @classmethod
     def get_batch(cls, keys):
@@ -558,16 +588,14 @@ class DynamoDBModel(object):
 
         get_batch *always* performs eventually consistent reads.
 
-        Please not that a batch can *not* read more than 100 items at once.
+        Objects loaded by this method are marked as coming from the DB. Hence
+        their initial state is saved in ``self._raw_data``.
 
         :param keys: iterable of keys. ex ``[(hash1, range1), (hash2, range2)]``
 
         """
-        if len(keys) > 100:
-            raise ValueError("Too many items to read in a single batch. Maximum is 100.")
+        table = ConnectionBorg().get_table(cls.__table__)
 
-        borg = ConnectionBorg()
-        table = borg.get_table(cls.__table__)
         # Convert all the keys to DynamoDB values.
         if cls.__range_key__:
             dynamo_keys = [
@@ -579,16 +607,11 @@ class DynamoDBModel(object):
         else:
             dynamo_keys = map(_python_to_dynamodb, keys)
 
-        batch_list = borg.new_batch_list()
-        batch_list.add_batch(table, dynamo_keys)
-
-        res = batch_list.submit()
+        res = table.batch_get_item(dynamo_keys)
 
         dblog.debug("Sent a batch get on table %s", cls.__table__)
 
-        return [
-            cls.from_dict(d) for d in res[u"Responses"][cls.__table__][u"Items"]
-        ]
+        return [cls._from_db_dict(d) for d in res]
 
     @classmethod
     def query(cls, hash_key_value, range_key_condition=None, consistent_read=False, reverse=False, limit=None):
@@ -603,6 +626,9 @@ class DynamoDBModel(object):
         with hash-only primary keys, cls.get(k) does the same thing cls.query(k)
         would.
 
+        Objects loaded by this method are marked as coming from the DB. Hence
+        their initial state is saved in ``self._raw_data``.
+
         :param hash_key_value: The hash key's value for all requested items.
 
         :param range_key_condition: A condition instance from
@@ -642,7 +668,7 @@ class DynamoDBModel(object):
 
         dblog.debug("Queried (%s, %s) on table %s", h_value, range_key_condition, cls.__table__)
 
-        return (cls.from_dict(d) for d in res)
+        return (cls._from_db_dict(d) for d in res)
 
     @classmethod
     def scan(cls, scan_filter=None):
@@ -654,6 +680,9 @@ class DynamoDBModel(object):
         Scan is a very expensive operation -- it doesn't use any indexes and will
         look through the entire table. As much as possible, you should avoid it.
 
+        Objects loaded by this method are marked as coming from the DB. Hence
+        their initial state is saved in ``self._raw_data``.
+
         :param scan_filter: A ``{attribute_name: condition}`` dict, where
             condition is a condition instance from ``boto.dynamodb.condition``.
 
@@ -667,14 +696,19 @@ class DynamoDBModel(object):
         dblog.debug("Scanned table %s with filter %s", cls.__table__, scan_filter)
 
         return (
-            cls.from_dict(d)
+            cls._from_db_dict(d)
             for d in res
             if d[hash_key_name] != MAGIC_KEY or cls.__schema__[hash_key_name] != autoincrement_int
         )
 
-    def to_db_dict(self):
+    def _to_db_dict(self):
         """Return a dict representation of the object according to the class's
         schema, suitable for direct storage in DynamoDB.
+
+        Direct use of this method should be avoided as much as possible but still
+        may be usefull for "deep copy".
+
+        Overload this method if you need a special serialization semantic
         """
         out = {}
         for name in type(self).__schema__:
@@ -693,8 +727,7 @@ class DynamoDBModel(object):
         (in particular, sets must be converted to lists), but types not
         suitable for DynamoDB (e.g. nested data structures) may be used.
 
-        Note that this method is never used for interaction with the database
-        (:meth:`to_db_dict` is).
+        Note that this method is never used for interaction with the database.
         """
         out = {}
         for name in self.__schema__:
@@ -730,11 +763,13 @@ class DynamoDBModel(object):
             max_hash_item.add_attribute(counter_key, 1)
             max_hash_item = max_hash_item.save(return_values='ALL_NEW')
             # We just reserved that value for the hash key
-            item[item.hash_key_name] = max_hash_item['Attributes'][counter_key]
+            hash_key = max_hash_item['Attributes'][counter_key]
+            item[item.hash_key_name] = hash_key
 
             try:
                 # Make sure this primary key was not 'stolen' by a direct DB access
                 item.put({item.hash_key_name: False})
+                dblog.debug("Saved autoinc (%s) in table %s", hash_key, item.table)
                 return
             except DynamoDBResponseError as e:
                 if e.error_code != "ConditionalCheckFailedException":
@@ -747,31 +782,40 @@ class DynamoDBModel(object):
         # This table auto-incr has been screwed up...
         raise MaxRetriesExceededError()
 
-    def save(self, allow_overwrite=True, expected_values=None):
+    def save(self, raise_on_conflict=False):
         """Save the object to the database.
 
         This method may be used both to insert a new object in the DB, or to
-        update an existing one (iff ``allow_overwrite == True``).
+        update an existing one (iff ``raise_on_conflict == False``).
 
-        :param allow_overwrite: If False, the method will only succeed if this
-            object's primary keys don't exist in the database (otherwise,
-            :class:`OverwriteError` is raised).
+        It also embeds the high level logic to avoid the 'lost update' syndrom.
+        Internally, it uses ``expected_values`` set to ``self._raw_data``
 
-        :param expected_values: dict of expected attribute values -- if any
-            one of these values doesn't match what is in the database
+        ``raise_on_conflict=True`` scenarios:
+
+        - **object from database**: Use ``self._raw_dict`` to generate ``expected_values``
+        - **new object**: ``self._raw_dict`` is empty, set ``allow_overwrite=True``
+        - **new object with autoinc**: flag has no effect
+        - **(accidentally) editing keys**: Use ``self._raw_dict`` to generate ``expected_values``, will catch overwrites and insertion to empty location
+
+        :param raise_on_conflict: flag to toggle overwrite protection -- if any
+            one of the original values doesn't match what is in the database
             (i.e. someone went ahead and modified the object in the DB behind
             your back), the operation fails and raises
-            :class:`ExpectedValueError`.
+            :class:`ConflictError`.
+
+        :raise ConflictError: Target object has changed between read and write operation
+        :raise OverwriteError: Saving a new object but it already existed
         """
-        if expected_values is None:
-            expected_values = {}
 
         cls = type(self)
+        expected_values = {}
+        allow_overwrite = True
         schema = cls.__schema__
         hash_key = cls.__hash_key__
         range_key = cls.__range_key__
         table = ConnectionBorg().get_table(cls.__table__)
-        item_data = self.to_db_dict()
+        item_data = self._to_db_dict()
         item = Item(table, attrs=item_data)
 
         # Detect magic elem manual overwrite
@@ -785,49 +829,83 @@ class DynamoDBModel(object):
             setattr(self, hash_key, item[hash_key])
         # Regular save
         else:
-            if allow_overwrite:
-                expected_db_values = {}
-                schema = type(self).__schema__
-                for (name, value) in expected_values.iteritems():
-                    db_value = _python_to_dynamodb(value)
-                    if db_value is None:
-                        # Empty strings/sets must be represented as missing values
-                        expected_db_values[name] = False
-                    else:
-                        expected_db_values[name] = db_value
-            else:
-                # Forbid overwrites: do a conditional write on
-                # "this hash_key doesn't exist"
-                expected_db_values = {hash_key: False}
-                if range_key:
-                    expected_db_values[range_key] = False
+            if raise_on_conflict:
+                if self._raw_data:
+                    expected_values = self._raw_data
+                    # Empty strings/sets must be represented as missing values
+                    for name in schema.iterkeys():
+                        if name not in expected_values:
+                            expected_values[name] = False
+                else:
+                    # Forbid overwrites: do a conditional write on
+                    # "this hash_key doesn't exist"
+                    allow_overwrite = False
+                    expected_values = {hash_key: False}
+                    if range_key:
+                        expected_values[range_key] = False
             try:
-                item.put(expected_db_values)
+                item.put(expected_values)
             except DynamoDBResponseError as e:
                 if e.error_code == "ConditionalCheckFailedException":
                     if allow_overwrite:
-                        # Failed conditional write
-                        raise ExpectedValueError(item)
+                        # Conflict detected
+                        raise ConflictError(item)
                     # Forbidden overwrite
                     raise OverwriteError(item)
                 # Unhandled exception
                 raise
 
-        hash_key_value = getattr(self, hash_key)
-        range_key_value = getattr(self, hash_key, None)
-        dblog.debug("Saved (%s, %s) in table %s allow_overwrite=%s", hash_key_value, range_key_value, cls.__table__, allow_overwrite)
+        # Update Raw_data to reflect DB state on success
+        self._raw_data = self._to_db_dict()
 
-    def delete(self):
-        """Delete the current object from the database."""
+        hash_key_value = getattr(self, hash_key)
+        range_key_value = getattr(self, range_key, None) if range_key else None
+        dblog.debug("Saved (%s, %s) in table %s raise_on_conflict=%s", hash_key_value, range_key_value, cls.__table__, raise_on_conflict)
+
+    def delete(self, raise_on_conflict=False):
+        """Delete the current object from the database.
+
+        If the Item has been edited before the ``delete`` command is issued and
+        ``raise_on_conflict=True`` then, :class:`ConflictError` is raised.
+
+        :param raise_on_conflict: flag to toggle overwrite protection -- if any
+            one of the original values doesn't match what is in the database
+            (i.e. someone went ahead and modified the object in the DB behind
+            your back), the operation fails and raises
+            :class:`ConflictError`.
+
+        :raise ConflictError: Target object has changed between read and write operation
+        """
         cls = type(self)
+        schema = cls.__schema__
+        expected_values = None
         hash_key_value = getattr(self, cls.__hash_key__)
+        h_value = _python_to_dynamodb(hash_key_value)
+
+        if raise_on_conflict:
+            if self._raw_data:
+                expected_values = self._raw_data
+                # Empty strings/sets must be represented as missing values
+                for name in schema.iterkeys():
+                    if name not in expected_values:
+                        expected_values[name] = False
+            else: #shortcut :D
+                raise ConflictError("Attempts to delete an object which has not yet been persited with raise_on_conflict=True")
+
         # Range key is only present in composite primary keys
         if cls.__range_key__:
             range_key_value = getattr(self, cls.__range_key__)
+            r_value = _python_to_dynamodb(range_key_value)
         else:
-            range_key_value = None
+            r_value = None
 
-        table = ConnectionBorg().get_table(cls.__table__)
-        Item(table, hash_key_value, range_key_value).delete()
+        try:
+            table = ConnectionBorg().get_table(cls.__table__)
+            Item(table, h_value, r_value).delete(expected_values)
+        except DynamoDBConditionalCheckFailedError, e:
+            raise ConflictError(e)
 
-        dblog.debug("Deleted (%s, %s) from table %s", hash_key_value, range_key_value, cls.__table__)
+        # Make sure any further save will be considered as *insertion*
+        self._raw_data = {}
+
+        dblog.debug("Deleted (%s, %s) from table %s", h_value, r_value, cls.__table__)
