@@ -6,11 +6,16 @@ import mock
 
 from dynamodb_mapper.model import (ConnectionBorg, DynamoDBModel,
     autoincrement_int, MaxRetriesExceededError, MAX_RETRIES,
-    ExpectedValueError, _python_to_dynamodb, _dynamodb_to_python, utc_tz,
-    _get_default_value, ThroughputError, SchemaError, MAGIC_KEY, OverwriteError)
+    ConflictError, _python_to_dynamodb, _dynamodb_to_python, utc_tz,
+    _get_default_value, SchemaError, MAGIC_KEY, OverwriteError, InvalidRegionError,)
 from boto.exception import DynamoDBResponseError
+from boto.dynamodb.exceptions import DynamoDBConditionalCheckFailedError
 import datetime
 
+class mocked_region(object):
+    def __init__(self, name):
+        self.name = name
+
 
 # Hash-only primary key
 class DoomEpisode(DynamoDBModel):
@@ -30,6 +35,7 @@ class DoomMap(DynamoDBModel):
     __schema__ = {
         "episode_id": int,
         "name": unicode,
+        "world": unicode
     }
 
 
@@ -164,10 +170,20 @@ def return_42():
 
 class TestConnectionBorg(unittest.TestCase):
     def setUp(self):
-        ConnectionBorg()._connections.clear()
+        ConnectionBorg._shared_state = {
+            "_aws_access_key_id": None,
+            "_aws_secret_access_key": None,
+            "_region": None,
+            "_connection": None,
+        }
 
     def tearDown(self):
-        ConnectionBorg()._connections.clear()
+        ConnectionBorg._shared_state = {
+            "_aws_access_key_id": None,
+            "_aws_secret_access_key": None,
+            "_region": None,
+            "_connection": None,
+        }
 
     def test_borgness(self):
         aws_access_key_id = "foo"
@@ -178,47 +194,150 @@ class TestConnectionBorg(unittest.TestCase):
 
         borg1.set_credentials(aws_access_key_id, aws_secret_access_key)
 
-        self.assertEquals(borg1._aws_access_key_id, aws_access_key_id)
-        self.assertEquals(borg1._aws_secret_access_key, aws_secret_access_key)
+        self.assertEqual(borg1._aws_access_key_id, aws_access_key_id)
+        self.assertEqual(borg1._aws_secret_access_key, aws_secret_access_key)
 
-        self.assertEquals(borg2._aws_access_key_id, aws_access_key_id)
-        self.assertEquals(borg2._aws_secret_access_key, aws_secret_access_key)
+        self.assertEqual(borg2._aws_access_key_id, aws_access_key_id)
+        self.assertEqual(borg2._aws_secret_access_key, aws_secret_access_key)
 
     @mock.patch("dynamodb_mapper.model.boto")
-    def test_get_table_nominal(self, m_boto):
+    def test_get_connection(self, m_boto):
+        CONNECTION = "JJG"
+        REGION = "La bas, tout est neuf et tout est sauvage."
+        KEY_ID = "foo"
+        KEY_VALUE = "bar"
+
+        m_boto.connect_dynamodb.return_value = "JJG"
+
+        borg1 = ConnectionBorg()
+        borg2 = ConnectionBorg()
+
+        borg1._region = REGION
+        borg1._aws_access_key_id = KEY_ID
+        borg1._aws_secret_access_key = KEY_VALUE
+
+        self.assertIsNone(borg1._connection)
+        self.assertEqual(CONNECTION, borg2._get_connection())
+        self.assertEqual(CONNECTION, borg1._get_connection())
+        self.assertEqual(CONNECTION, borg1._connection)
+
+        m_boto.connect_dynamodb.assert_called_once_with(
+                aws_access_key_id=KEY_ID,
+                aws_secret_access_key=KEY_VALUE,
+                region=REGION,
+                )
+
+    @mock.patch("dynamodb_mapper.model.boto")
+    def test_set_region_valid(self, m_boto):
+        # Make sure internal state is set and shared
+        m_regions = [
+            mocked_region('us-east-1'),
+            mocked_region('eu-west-1'),
+        ]
+
+        m_boto.dynamodb.regions.return_value = m_regions
+
+        borg1 = ConnectionBorg()
+        borg2 = ConnectionBorg()
+
+        self.assertIs(None, borg1._region)
+        self.assertIs(None, borg2._region)
+
+        borg1.set_region("eu-west-1")
+
+        self.assertIs(m_regions[1], borg1._region)
+        self.assertIs(m_regions[1], borg2._region)
+
+    @mock.patch("dynamodb_mapper.model.boto")
+    def test_set_region_invalid(self, m_boto):
+        # Make sure internal state is set and shared
+        m_regions = [
+            mocked_region('us-east-1'),
+            mocked_region('eu-west-1'),
+        ]
+
+        m_boto.dynamodb.regions.return_value = m_regions
+
+        borg1 = ConnectionBorg()
+        borg2 = ConnectionBorg()
+
+        self.assertIs(None, borg1._region)
+        self.assertIs(None, borg2._region)
+
+        self.assertRaises(
+            InvalidRegionError,
+            borg1.set_region,
+            "moon-hidden-1",
+        )
+
+        self.assertIs(None, borg1._region)
+        self.assertIs(None, borg2._region)
+
+
+    @mock.patch("dynamodb_mapper.model.boto")
+    def test_set_credentials(self, m_boto):
+        # Make sure internal state is set and shared
         m_get_table = m_boto.connect_dynamodb.return_value.get_table
 
         aws_access_key_id = "foo"
         aws_secret_access_key = "bar"
         table_name = "foo"
-        table_name2 = "bar"
 
-        ConnectionBorg().set_credentials(aws_access_key_id, aws_secret_access_key)
+        borg1 = ConnectionBorg()
+        borg2 = ConnectionBorg()
+
+        borg1.set_credentials(aws_access_key_id, aws_secret_access_key)
+
+        self.assertIs(borg1._aws_access_key_id ,aws_access_key_id)
+        self.assertIs(borg2._aws_access_key_id ,aws_access_key_id)
+
+        self.assertIs(borg1._aws_secret_access_key ,aws_secret_access_key)
+        self.assertIs(borg2._aws_secret_access_key ,aws_secret_access_key)
+
+    @mock.patch("dynamodb_mapper.model.boto")
+    def test_get_table_default(self, m_boto):
+        m_get_table = m_boto.connect_dynamodb.return_value.get_table
+
+        aws_access_key_id = "foo"
+        aws_secret_access_key = "bar"
+        table_name = "foo"
+
+        borg1 = ConnectionBorg()
+
+        borg1._aws_access_key_id = aws_access_key_id
+        borg1._aws_secret_access_key = aws_secret_access_key
 
         ConnectionBorg().get_table(table_name)
         m_get_table.assert_called_with(table_name)
 
-        ConnectionBorg().get_table(table_name2)
-        m_get_table.assert_called_with(table_name2)
-
-        # Should be called only once per thread even when getting multiple tables.
-        m_boto.connect_dynamodb.assert_called_once_with(
+        m_boto.connect_dynamodb.assert_called_with(
             aws_access_key_id=aws_access_key_id,
-            aws_secret_access_key=aws_secret_access_key
+            aws_secret_access_key=aws_secret_access_key,
+            region=None,
         )
 
     @mock.patch("dynamodb_mapper.model.boto")
-    def test_new_batch_list_nominal(self, m_boto):
-        m_new_batch_list = m_boto.connect_dynamodb.return_value.new_batch_list
-        ConnectionBorg().new_batch_list()
+    def test_get_table_eu_region(self, m_boto):
+        m_get_table = m_boto.connect_dynamodb.return_value.get_table
+        m_region = mocked_region('eu-west-1')
 
-        m_new_batch_list.assert_called_once()
+        aws_access_key_id = "foo"
+        aws_secret_access_key = "bar"
+        table_name = "foo"
 
-    def test_create_table_bad_throughput(self):
-        self.assertRaises(
-            ThroughputError,
-            ConnectionBorg().create_table,
-            DoomEpisode, 2, 5, False
+        borg1 = ConnectionBorg()
+
+        borg1._region = m_region
+        borg1._aws_access_key_id = aws_access_key_id
+        borg1._aws_secret_access_key = aws_secret_access_key
+
+        ConnectionBorg().get_table(table_name)
+        m_get_table.assert_called_with(table_name)
+
+        m_boto.connect_dynamodb.assert_called_with(
+            aws_access_key_id=aws_access_key_id,
+            aws_secret_access_key=aws_secret_access_key,
+            region=m_region,
         )
 
     def test_create_table_no_name(self):
@@ -334,42 +453,42 @@ class TestConnectionBorg(unittest.TestCase):
 
 class TestTypeConversions(unittest.TestCase):
     def test_python_to_dynamodb_number(self):
-        self.assertEquals(0, _python_to_dynamodb(0))
-        self.assertEquals(0.0, _python_to_dynamodb(0.0))
-        self.assertEquals(10, _python_to_dynamodb(10))
-        self.assertEquals(10.0, _python_to_dynamodb(10.0))
-        self.assertEquals(0, _python_to_dynamodb(False))
-        self.assertEquals(1, _python_to_dynamodb(True))
+        self.assertEqual(0, _python_to_dynamodb(0))
+        self.assertEqual(0.0, _python_to_dynamodb(0.0))
+        self.assertEqual(10, _python_to_dynamodb(10))
+        self.assertEqual(10.0, _python_to_dynamodb(10.0))
+        self.assertEqual(0, _python_to_dynamodb(False))
+        self.assertEqual(1, _python_to_dynamodb(True))
 
     def test_dynamodb_to_python_number(self):
-        self.assertEquals(0, _dynamodb_to_python(int, 0))
-        self.assertEquals(0.0, _dynamodb_to_python(float, 0.0))
-        self.assertEquals(10, _dynamodb_to_python(int, 10))
-        self.assertEquals(10.0, _dynamodb_to_python(float, 10.0))
-        self.assertEquals(False, _dynamodb_to_python(bool, 0))
-        self.assertEquals(True, _dynamodb_to_python(bool, 1))
+        self.assertEqual(0, _dynamodb_to_python(int, 0))
+        self.assertEqual(0.0, _dynamodb_to_python(float, 0.0))
+        self.assertEqual(10, _dynamodb_to_python(int, 10))
+        self.assertEqual(10.0, _dynamodb_to_python(float, 10.0))
+        self.assertEqual(False, _dynamodb_to_python(bool, 0))
+        self.assertEqual(True, _dynamodb_to_python(bool, 1))
 
     def test_python_to_dynamodb_unicode(self):
-        self.assertEquals(u"hello", _python_to_dynamodb(u"hello"))
+        self.assertEqual(u"hello", _python_to_dynamodb(u"hello"))
         # Empty strings become missing attributes
         self.assertIsNone(_python_to_dynamodb(u""))
 
     def test_dynamodb_to_python_unicode(self):
-        self.assertEquals(u"hello", _dynamodb_to_python(unicode, u"hello"))
+        self.assertEqual(u"hello", _dynamodb_to_python(unicode, u"hello"))
         # Missing attributes become None
         self.assertIsNone(_dynamodb_to_python(unicode, None))
 
     def test_python_to_dynamodb_set(self):
-        self.assertEquals(set([1, 2, 3]), _python_to_dynamodb(set([1, 2, 3])))
-        self.assertEquals(
+        self.assertEqual(set([1, 2, 3]), _python_to_dynamodb(set([1, 2, 3])))
+        self.assertEqual(
             set(["foo", "bar", "baz"]),
             _python_to_dynamodb(set(["foo", "bar", "baz"])))
         # Empty sets become missing attributes
         self.assertIsNone(_python_to_dynamodb(set()))
 
     def test_dynamodb_to_python_set(self):
-        self.assertEquals(set([1, 2, 3]), _dynamodb_to_python(set, set([1, 2, 3])))
-        self.assertEquals(
+        self.assertEqual(set([1, 2, 3]), _dynamodb_to_python(set, set([1, 2, 3])))
+        self.assertEqual(
             set(["foo", "bar", "baz"]),
             _dynamodb_to_python(set, set(["foo", "bar", "baz"])))
         # Empty sets become None
@@ -381,10 +500,10 @@ class TestTypeConversions(unittest.TestCase):
             {'name': 'punch', 'damage': 5}
         ]
 
-        self.assertEquals(
+        self.assertEqual(
             json.dumps(attacks, sort_keys=True),
             _python_to_dynamodb(attacks))
-        self.assertEquals("[]", _python_to_dynamodb([]))
+        self.assertEqual("[]", _python_to_dynamodb([]))
 
     def test_dynamodb_to_python_list(self):
         attacks = [
@@ -392,10 +511,10 @@ class TestTypeConversions(unittest.TestCase):
             {'name': 'punch', 'damage': 5}
         ]
 
-        self.assertEquals(
+        self.assertEqual(
             attacks,
             _dynamodb_to_python(list, json.dumps(attacks, sort_keys=True)))
-        self.assertEquals([], _dynamodb_to_python(list, "[]"))
+        self.assertEqual([], _dynamodb_to_python(list, "[]"))
 
     def test_python_to_dynamodb_dict(self):
         monsters = {
@@ -407,10 +526,10 @@ class TestTypeConversions(unittest.TestCase):
             "cyberdemon": []
         }
 
-        self.assertEquals(
+        self.assertEqual(
             json.dumps(monsters, sort_keys=True),
             _python_to_dynamodb(monsters))
-        self.assertEquals("{}", _python_to_dynamodb({}))
+        self.assertEqual("{}", _python_to_dynamodb({}))
 
     def test_dynamodb_to_python_dict(self):
         monsters = {
@@ -422,20 +541,20 @@ class TestTypeConversions(unittest.TestCase):
             "cyberdemon": []
         }
 
-        self.assertEquals(
+        self.assertEqual(
             monsters,
             _dynamodb_to_python(dict, json.dumps(monsters, sort_keys=True)))
-        self.assertEquals({}, _dynamodb_to_python(dict, "{}"))
+        self.assertEqual({}, _dynamodb_to_python(dict, "{}"))
 
     def test_dynamodb_to_python_datetime(self):
-        self.assertEquals(
-            datetime.datetime(2012, 05, 31, 12, 0, 0, tzinfo=utc_tz),
+        self.assertEqual(
+            datetime.datetime(2012, 05, 31, 12, 0, 0, 42, tzinfo=utc_tz),
             _dynamodb_to_python(datetime.datetime,
-                                "2012-05-31T12:00:00.000000+00:00"))
-        self.assertEquals(
-            datetime.datetime(2010, 11, 1, 4, 0, 0, tzinfo=utc_tz),
+                                "2012-05-31T12:00:00.000042+00:00"))
+        self.assertEqual(
+            datetime.datetime(2010, 11, 1, 4, 0, 0, 13, tzinfo=utc_tz),
             _dynamodb_to_python(datetime.datetime,
-                                "2010-11-01T04:00:00.000000Z"))
+                                "2010-11-01T04:00:00.000013Z"))
 
     def test_dynamodb_to_python_default(self):
         self.assertIsNone(_dynamodb_to_python(datetime.datetime, None))
@@ -447,7 +566,7 @@ class TestTypeConversions(unittest.TestCase):
             _dynamodb_to_python, datetime.datetime, "2012-05-31T12:00:00.000000")
 
     def test_python_to_dynamodb_datetime(self):
-        self.assertEquals(
+        self.assertEqual(
             "2012-05-31T12:00:00.000000+00:00",
             _python_to_dynamodb(datetime.datetime(2012, 05, 31, 12, 0, 0, tzinfo=utc_tz)))
 
@@ -462,23 +581,24 @@ class TestGetDefaultValue(unittest.TestCase):
     def test_datetime(self, m_datetime):
         # Plain default date should be now
         m_datetime.now.return_value = now = datetime.datetime.now(tz=utc_tz)
-        self.assertEquals(
+        self.assertEqual(
             now,
             _get_default_value(m_datetime))
+        self.assertEqual(utc_tz.tzname(now), "UTC")
 
         # default date with supplied fallback
         fallback = datetime.datetime(2012, 05, 31, 12, 0, 0, tzinfo=utc_tz)
-        self.assertEquals(
+        self.assertEqual(
             fallback,
             _get_default_value(datetime.datetime, fallback))
 
     def test_int(self):
         # Plain default int
-        self.assertEquals(0, _get_default_value(int))
+        self.assertEqual(0, _get_default_value(int))
         # Default int with fallback
-        self.assertEquals(42, _get_default_value(int, 42))
+        self.assertEqual(42, _get_default_value(int, 42))
         # Default int with callable fallback
-        self.assertEquals(42, _get_default_value(int, return_42))
+        self.assertEqual(42, _get_default_value(int, return_42))
 
     def test_incompatible_types_crash_constant(self):
         # Default unicode with int fallback should fail
@@ -499,10 +619,10 @@ class TestGetDefaultValue(unittest.TestCase):
 
 class TestDynamoDBModel(unittest.TestCase):
     def setUp(self):
-        ConnectionBorg()._connections.clear()
+        ConnectionBorg()._connection = None
 
     def tearDown(self):
-        ConnectionBorg()._connections.clear()
+        ConnectionBorg()._connection = None
 
     def test_to_json_dict(self):
         testdate = datetime.datetime(2012, 5, 31, 12, 0, 0, tzinfo=utc_tz)
@@ -514,63 +634,69 @@ class TestDynamoDBModel(unittest.TestCase):
         d.date = testdate
 
         d_dict = d.to_json_dict()
-        self.assertEquals(d_dict["id"], 42)
-        self.assertEquals(d_dict["set"], sorted(testSet))
-        self.assertEquals(d_dict["date"], testdate.astimezone(utc_tz).isoformat())
+        self.assertEqual(d_dict["id"], 42)
+        self.assertEqual(d_dict["set"], sorted(testSet))
+        self.assertEqual(d_dict["date"], testdate.astimezone(utc_tz).isoformat())
 
     def test_build_default_values(self):
         d = DoomEpisode()
-        self.assertEquals(d.id, 0)
-        self.assertEquals(d.name, u"")
+        self.assertEqual(d.id, 0)
+        self.assertEqual(d.name, u"")
 
-        # Using to_json_dict because to_db_dict omits empty values
+        # Using to_json_dict because _to_db_dict omits empty values
         d_dict = d.to_json_dict()
-        self.assertEquals(d_dict["id"], d.id)
-        self.assertEquals(d_dict["name"], d.name)
+        self.assertEqual(d_dict["id"], d.id)
+        self.assertEqual(d_dict["name"], d.name)
 
     @mock.patch("dynamodb_mapper.model._get_default_value")
     def test_build_default_values_with_defaults(self, m_defaulter):
         m_defaulter.return_value = u"weak"
         d = PlayerStrength()
 
-        m_defaulter.assert_called_with(unicode, u"weak")
-        self.assertEquals(d.strength, u"weak")
+        m_defaulter.assert_called_with(unicode, u"weak", name='strength')
+        self.assertEqual(d.strength, u"weak")
 
     def test_init_from_args(self):
         d = DoomEpisode(id=1, name=u"Knee-deep in the Dead")
-        self.assertEquals(1, d.id)
-        self.assertEquals(u"Knee-deep in the Dead", d.name)
+        self.assertEqual({}, d._raw_data)
+        self.assertEqual(1, d.id)
+        self.assertEqual(u"Knee-deep in the Dead", d.name)
 
     def test_build_from_dict(self):
+        #FIXME: can it be removed as this feature is implicitly tested elsewhere ?
         d_dict = {"id": 1, "name": "Knee-deep in the Dead"}
-        d = DoomEpisode.from_dict(d_dict)
-        self.assertEquals(d_dict["id"], d.id)
-        self.assertEquals(d_dict["name"], d.name)
+        d = DoomEpisode._from_db_dict(d_dict)
+        self.assertEqual(d_dict, d._raw_data)
+        self.assertEqual(d_dict["id"], d.id)
+        self.assertEqual(d_dict["name"], d.name)
 
     def test_build_from_dict_missing_attrs(self):
-        d = DoomCampaign.from_dict({})
-        self.assertEquals(d.id, 0)
-        self.assertEquals(d.name, u"")
-        self.assertEquals(d.cheats, set())
+        #FIXME: can it be removed as this feature is implicitly tested elsewhere ?
+        d = DoomCampaign._from_db_dict({})
+        self.assertEqual(d.id, 0)
+        self.assertEqual(d.name, u"")
+        self.assertEqual(d.cheats, set())
 
     def test_build_from_dict_json_list(self):
+        #FIXME: can it be removed as this feature is implicitly tested elsewhere ?
         attacks = [
             {'name': 'fireball', 'damage': 10},
             {'name': 'punch', 'damage': 5}
         ]
 
-        m = DoomMonster.from_dict({
+        m = DoomMonster._from_db_dict({
             "id": 1,
             "attacks": json.dumps(attacks, sort_keys=True)
         })
 
-        self.assertEquals(m.attacks, attacks)
+        self.assertEqual(m.attacks, attacks)
 
         # Test default values
-        m2 = DoomMonster.from_dict({"id": 1})
-        self.assertEquals(m2.attacks, [])
+        m2 = DoomMonster._from_db_dict({"id": 1})
+        self.assertEqual(m2.attacks, [])
 
     def test_build_from_dict_json_dict(self):
+        #FIXME: can it be removed as this feature is implicitly tested elsewhere ?
         monsters = {
             "cacodemon": [
                 {"x": 10, "y": 20},
@@ -580,16 +706,44 @@ class TestDynamoDBModel(unittest.TestCase):
             "cyberdemon": []
         }
 
-        e = DoomMonsterMap.from_dict({
+        e = DoomMonsterMap._from_db_dict({
             "map_id": 1,
             "monsters": json.dumps(monsters, sort_keys=True)
         })
-        self.assertEquals(e.monsters, monsters)
+        self.assertEqual(e.monsters, monsters)
 
-        e2 = DoomMonsterMap.from_dict({"map_id": 1})
-        self.assertEquals(e2.monsters, {})
+        e2 = DoomMonsterMap._from_db_dict({"map_id": 1})
+        self.assertEqual(e2.monsters, {})
+
+    def test_build_from_db_dict(self):
+        #FIXME: can it be removed as this feature is implicitly tested elsewhere ?
+        d_dict = {"id": 1, "name": "Knee-deep in the Dead"}
+        d = DoomEpisode._from_db_dict(d_dict)
+        self.assertEqual(d_dict, d._raw_data)
+
+    def test_build_from_db_dict_json_dict(self):
+        #FIXME: can it be removed as this feature is implicitly tested elsewhere ?
+        monsters = {
+            "cacodemon": [
+                {"x": 10, "y": 20},
+                {"x": 10, "y": 30}
+            ],
+            "imp": [],
+            "cyberdemon": []
+        }
+
+        raw_dict = {
+            "map_id": 1,
+            "monsters": json.dumps(monsters, sort_keys=True)
+        }
+
+        e = DoomMonsterMap._from_db_dict(raw_dict)
+        self.assertEqual(e.monsters, monsters)
+        self.assertEqual(raw_dict["map_id"], e._raw_data["map_id"])
+        self.assertEqual(raw_dict["monsters"], e._raw_data["monsters"])
 
     def test_to_db_dict_json_list(self):
+        #FIXME: can it be removed as this feature is implicitly tested elsewhere ?
         attacks = [
             {'name': 'fireball', 'damage': 10},
             {'name': 'punch', 'damage': 5}
@@ -599,12 +753,13 @@ class TestDynamoDBModel(unittest.TestCase):
         m.id = 1
         m.attacks = attacks
 
-        d = m.to_db_dict()
+        d = m._to_db_dict()
 
-        self.assertEquals(d["id"], 1)
-        self.assertEquals(d["attacks"], json.dumps(attacks, sort_keys=True))
+        self.assertEqual(d["id"], 1)
+        self.assertEqual(d["attacks"], json.dumps(attacks, sort_keys=True))
 
     def test_to_db_dict_json_dict(self):
+        #FIXME: can it be removed as this feature is implicitly tested elsewhere ?
         monsters = {
             "cacodemon": [
                 {"x": 10, "y": 20},
@@ -618,10 +773,11 @@ class TestDynamoDBModel(unittest.TestCase):
         e.map_id = 1
         e.monsters = monsters
 
-        d = e.to_db_dict()
+        d = e._to_db_dict()
+
+        self.assertEqual(d["map_id"], 1)
+        self.assertEqual(d["monsters"], json.dumps(monsters, sort_keys=True))
 
-        self.assertEquals(d["map_id"], 1)
-        self.assertEquals(d["monsters"], json.dumps(monsters, sort_keys=True))
 
     @mock.patch("dynamodb_mapper.model.Item")
     @mock.patch("dynamodb_mapper.model.boto")
@@ -683,7 +839,8 @@ class TestDynamoDBModel(unittest.TestCase):
         d_text = "2012-05-31T12:00:00.000000+00:00"
         m_table.get_item.return_value = {"datetime": d_text}
 
-        Patch.get(d)
+        p = Patch.get(d)
+        self.assertEqual(d_text, p._raw_data["datetime"])
         m_table.get_item.assert_called_once_with(
             hash_key=d_text, range_key=None, consistent_read=False)
 
@@ -699,7 +856,9 @@ class TestDynamoDBModel(unittest.TestCase):
             "datetime": d_text
         }
 
-        GameReport.get(players, d)
+        r = GameReport.get(players, d)
+        self.assertEqual(d_text, r._raw_data["datetime"])
+        self.assertEqual(players_text, r._raw_data["player_ids"])
         m_table.get_item.assert_called_once_with(
             hash_key=players_text, range_key=d_text, consistent_read=False)
 
@@ -753,7 +912,7 @@ class TestDynamoDBModel(unittest.TestCase):
         l.save()
 
         m_item.return_value.__setitem__.assert_called_with("id", 3)
-        self.assertEquals(m_item.return_value.__setitem__.call_count, 2)
+        self.assertEqual(m_item.return_value.__setitem__.call_count, 2)
 
     @mock.patch("dynamodb_mapper.model.Item")
     @mock.patch("dynamodb_mapper.model.boto")
@@ -771,7 +930,7 @@ class TestDynamoDBModel(unittest.TestCase):
         l.text = "Everybody's dead, Dave."
         self.assertRaises(MaxRetriesExceededError, l.save)
 
-        self.assertEquals(m_item_instance.put.call_count, MAX_RETRIES)
+        self.assertEqual(m_item_instance.put.call_count, MAX_RETRIES)
 
     @mock.patch("dynamodb_mapper.model.Item")
     @mock.patch("dynamodb_mapper.model.boto")
@@ -800,40 +959,57 @@ class TestDynamoDBModel(unittest.TestCase):
 
     @mock.patch("dynamodb_mapper.model.Item")
     @mock.patch("dynamodb_mapper.model.boto")
-    def test_save_expected_values(self, m_boto, m_item):
+    def test_save_raise_on_conflict(self, m_boto, m_item):
         m_item_instance = m_item.return_value
+        error = DynamoDBResponseError(
+            None, None, {"__type": "ConditionalCheckFailedException"})
+        m_item_instance.put.side_effect = error
 
         name = "Knee-deep in the Dead"
         cheats = set(["iddqd", "idkfa"])
 
         c = DoomCampaign()
-        c.id = 0
+
+        c._raw_data["id"] = 1
+        c._raw_data["name"] = name
+        #simulate empty field in DB
+        #c._raw_data["cheats"] = cheats
+
+        c.id = 1
         c.name = name
         c.cheats = cheats
 
-        c.save(expected_values={"id": 1, "name": name, "cheats": cheats})
-        m_item_instance.put.assert_called_with({"id": 1, "name": name, "cheats": cheats})
-
+        self.assertRaises(ConflictError, c.save, raise_on_conflict=True)
+        m_item_instance.put.assert_called_with({"id": 1, "name": name, "cheats": False})
 
     @mock.patch("dynamodb_mapper.model.Item")
     @mock.patch("dynamodb_mapper.model.boto")
-    def test_save_expected_values_boolean(self, m_boto, m_item):
+    def test_save_raise_on_conflict_boolean(self, m_boto, m_item):
         m_item_instance = m_item.return_value
+        error = DynamoDBResponseError(
+            None, None, {"__type": "ConditionalCheckFailedException"})
+        m_item_instance.put.side_effect = error
 
         name = "Knee-deep in the Dead"
         completed = False
 
         c = DoomCampaignStatus()
+
+        c._raw_data["id"] = 1
+        c._raw_data["name"] = name
+        c._raw_data["completed"] = int(False) # DynamoDB stores int and unicode only
+
         c.id = 0
         c.name = name
         c.completed = completed
 
-        c.save(expected_values={"id": 1, "name": name, "completed": completed})
+        self.assertRaises(ConflictError, c.save, raise_on_conflict=True)
         m_item_instance.put.assert_called_with({"id": 1, "name": name, "completed": 0})
 
     @mock.patch("dynamodb_mapper.model.Item")
     @mock.patch("dynamodb_mapper.model.boto")
     def test_save_no_overwrite_composite_fails(self, m_boto, m_item):
+        """Save raised OverwriteError when new object has the same ID"""
         m_item_instance = m_item.return_value
         error = DynamoDBResponseError(
             None, None, {"__type": "ConditionalCheckFailedException"})
@@ -845,7 +1021,7 @@ class TestDynamoDBModel(unittest.TestCase):
 
         self.assertRaises(
             OverwriteError,
-            c.save, allow_overwrite=False
+            c.save, raise_on_conflict=True
         )
         m_item_instance.put.assert_called_with({"episode_id": False, "name": False})
 
@@ -863,107 +1039,85 @@ class TestDynamoDBModel(unittest.TestCase):
 
         self.assertRaises(DynamoDBResponseError, c.save)
 
-    @mock.patch("dynamodb_mapper.model.Item")
-    @mock.patch("dynamodb_mapper.model.boto")
-    def test_save_expected_values_false_values(self, m_boto, m_item):
-        m_item_instance = m_item.return_value
-        error = DynamoDBResponseError(
-            None, None, {"__type": "ConditionalCheckFailedException"})
-        m_item_instance.put.side_effect = error
-
-        name = "Knee-deep in the Dead"
-        cheats = set(["iddqd", "idkfa"])
-
-        c = DoomCampaign()
-        c.id = 0
-        c.name = name
-        c.cheats = cheats
-
-        self.assertRaises(
-            ExpectedValueError,
-            c.save, expected_values={"id": 0, "name": "", "cheats": set()}
-        )
-
-    @mock.patch("dynamodb_mapper.model.Item")
-    @mock.patch("dynamodb_mapper.model.boto")
-    def test_save_expected_values_conditional_check_failed(self, m_boto, m_item):
-        m_item_instance = m_item.return_value
-        m_item_instance.put.side_effect
-
-        name = "Knee-deep in the Dead"
-        cheats = set(["iddqd", "idkfa"])
-
-        c = DoomCampaign()
-        c.id = 0
-        c.name = name
-        c.cheats = cheats
-
-        c.save(expected_values={"id": 1, "name": name, "cheats": cheats})
-
     @mock.patch("dynamodb_mapper.model.boto")
     def test_get_batch_hash_key(self, m_boto):
-        m_batch_list = m_boto.connect_dynamodb.return_value.new_batch_list.return_value
-        m_table = m_boto.connect_dynamodb.return_value.get_table.return_value
-        # Mock empty answer as we only observe the way the request is build
-        m_batch_list.submit.return_value = {
-            u"Responses": {
-                DoomEpisode.__table__: {
-                    u"Items": []
-                }
-            }
-        }
+        # paging function/generator is done by real db tests
+        KEYS = range(2)
+        DATA = [{
+                    u"id": 0,
+                    u"name": u"The level with cute cats",
+                },
+                {
+                    u"id": 1,
+                    u"name": u"The level with horrible monsters",
+                }]
 
-        DoomEpisode.get_batch(range(2));
-        m_batch_list.add_batch.assert_called_with(m_table, [0,1])
-        assert m_batch_list.submit.called
+        m_table = m_boto.connect_dynamodb.return_value.get_table.return_value
+        m_table.batch_get_item.return_value = DATA
+
+        res = DoomEpisode.get_batch(KEYS);
+
+        m_table.batch_get_item.assert_called_with(KEYS)
+        self.assertEqual(2, len(res))
+        self.assertEqual(DATA[0][u'id'], res[0].id)
+        self.assertEqual(DATA[0][u'name'], res[0].name)
+        self.assertEqual(DATA[1][u'id'], res[1].id)
+        self.assertEqual(DATA[1][u'name'], res[1].name)
 
     @mock.patch("dynamodb_mapper.model.boto")
     def test_get_batch_hash_range_key(self, m_boto):
-        m_batch_list = m_boto.connect_dynamodb.return_value.new_batch_list.return_value
+        KEYS = [(1, u"The level with cute cats"),
+                (2, u"The level with horrible monsters")]
+        DATA = [{
+                    u"episode_id": 1,
+                    u"name": u"The level with cute cats",
+                    u"world": u"Heaven",
+                },
+                {
+                    u"episode_id": 2,
+                    u"name": u"The level with horrible monsters",
+                    u"world": u"Hell",
+                }]
+
         m_table = m_boto.connect_dynamodb.return_value.get_table.return_value
-        # Mock empty answer as we only observe the way the request is build
-        m_batch_list.submit.return_value = {
-            u"Responses": {
-                DoomMap.__table__: {
-                    u"Items": []
-                }
-            }
-        }
+        m_table.batch_get_item.return_value = DATA
 
-        DoomMap.get_batch([(0, u"level1"), (1, u"level2")]);
-        m_batch_list.add_batch.assert_called_with(m_table, [(0, u'level1'), (1, u'level2')])
-        assert m_batch_list.submit.called
+        res = DoomMap.get_batch(KEYS)
 
-    def test_get_batch_over_100(self):
-        self.assertRaises(
-            ValueError,
-            DoomEpisode.get_batch,
-            range(101)
-        )
+        m_table.batch_get_item.assert_called_with(KEYS)
+        self.assertEqual(2, len(res))
+        self.assertEqual(DATA[0][u'episode_id'], res[0].episode_id)
+        self.assertEqual(DATA[0][u'name'], res[0].name)
+        self.assertEqual(DATA[0][u'world'], res[0].world)
+        self.assertEqual(DATA[1][u'episode_id'], res[1].episode_id)
+        self.assertEqual(DATA[1][u'name'], res[1].name)
+        self.assertEqual(DATA[1][u'world'], res[1].world)
 
     @mock.patch("dynamodb_mapper.model.ConnectionBorg.get_table")
     def test_query(self, m_get_table):
         m_table = m_get_table.return_value
         m_table.query.return_value = []
+        # FIXME: make sure _raw_data is filled
 
         from boto.dynamodb import condition
         DoomMap.query(1, condition.BEGINS_WITH(u"level"), True)
 
-        m_table.query.assert_called_with(1, condition.BEGINS_WITH(u"level"), consistent_read=True, scan_index_forward=True)
+        m_table.query.assert_called_with(1, condition.BEGINS_WITH(u"level"), consistent_read=True, scan_index_forward=True, max_results=None)
 
     @mock.patch("dynamodb_mapper.model.ConnectionBorg.get_table")
-    def test_query_reverse(self, m_get_table):
+    def test_query_reverse_limit(self, m_get_table):
         m_table = m_get_table.return_value
         m_table.query.return_value = []
 
         from boto.dynamodb import condition
-        DoomMap.query(1, condition.BEGINS_WITH(u"level"), True, reverse=True)
+        DoomMap.query(1, condition.BEGINS_WITH(u"level"), True, reverse=True, limit=42)
 
-        m_table.query.assert_called_with(1, condition.BEGINS_WITH(u"level"), consistent_read=True, scan_index_forward=False)
+        m_table.query.assert_called_with(1, condition.BEGINS_WITH(u"level"), consistent_read=True, scan_index_forward=False, max_results=42)
 
     @mock.patch("dynamodb_mapper.model.ConnectionBorg.get_table")
     def test_scan(self, m_get_table):
         # FIXME: test the generator. autoinc keys should be hidden from the results
+        # FIXME: make sure _raw_data is filled
         m_table = m_get_table.return_value
         m_table.scan.return_value = []
 
@@ -981,13 +1135,20 @@ class TestDynamoDBModel(unittest.TestCase):
         m_item_instance = m_item.return_value
         m_table = m_boto.connect_dynamodb.return_value.get_table.return_value
 
-        # Tests with a hash key elem
-        d = DoomEpisode()
-        d.id = "42"
+        python_date = datetime.datetime(2012, 05, 31, 12, 0, 0, tzinfo=utc_tz)
+        dynamodb_date = _python_to_dynamodb(python_date)
+
+        raw_data = {
+            u"datetime":dynamodb_date
+        }
+
+        # Simulate obj from DB
+        d = Patch._from_db_dict(raw_data)
 
         d.delete()
 
-        m_item.assert_called_once_with(m_table, "42", None)
+        self.assertEqual(d._raw_data, {})
+        m_item.assert_called_once_with(m_table, dynamodb_date, None)
         assert m_item_instance.delete.called
 
     @mock.patch("dynamodb_mapper.model.boto")
@@ -996,12 +1157,63 @@ class TestDynamoDBModel(unittest.TestCase):
         m_item_instance = m_item.return_value
         m_table = m_boto.connect_dynamodb.return_value.get_table.return_value
 
-        # Tests with a composite key elem
-        d = DoomMap()
-        d.episode_id = "42"
-        d.name = "level super geek"
+        raw_data = {
+            u"episode_id": 42,
+            u"name": "level super geek"
+        }
+
+        # Simulate obj from DB, composite keys
+        d = DoomMap._from_db_dict(raw_data)
 
         d.delete()
 
-        m_item.assert_called_once_with(m_table, "42", "level super geek")
+        self.assertEqual(d._raw_data, {})
+        m_item.assert_called_once_with(m_table, 42, "level super geek")
         assert m_item_instance.delete.called
+
+    @mock.patch("dynamodb_mapper.model.boto")
+    @mock.patch("dynamodb_mapper.model.Item")
+    def test_delete_new_object_roc(self, m_item, m_boto):
+        m_item_instance = m_item.return_value
+        m_table = m_boto.connect_dynamodb.return_value.get_table.return_value
+
+        # Create a new object
+        d = DoomEpisode()
+        d.id = "42"
+
+        self.assertRaises(
+            ConflictError,
+            d.delete,
+            raise_on_conflict=True
+        )
+
+        self.assertEqual(d._raw_data, {}) #still "new"
+
+    @mock.patch("dynamodb_mapper.model.boto")
+    @mock.patch("dynamodb_mapper.model.Item")
+    def test_delete_object_roc(self, m_item, m_boto):
+        m_item_instance = m_item.return_value
+        m_table = m_boto.connect_dynamodb.return_value.get_table.return_value
+        m_item_instance.delete.side_effect = DynamoDBConditionalCheckFailedError(404, "mock")
+
+        raw_data = {
+            u"episode_id": 42,
+            u"name": "level super geek"
+        }
+
+        # Simulate obj from DB
+        d = DoomMap._from_db_dict(raw_data)
+
+        self.assertRaises(
+            ConflictError,
+            d.delete,
+            raise_on_conflict=True
+        )
+
+        m_item_instance.delete.assert_called_with({
+            u'name': 'level super geek',
+            u'episode_id': 42,
+            u'world': False
+        })
+
+        self.assertEqual(d._raw_data, raw_data) #no change
